# Ksplice Uptrack API client library for Python
#
# Copyright (C) 2010 Ksplice, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import cjson
import urllib
import urlparse
import posixpath
import httplib
import ConfigParser

from ksplice.uptrack.errors import UptrackError, AuthenticationError, UsageError

API_URL         = "https://uptrack.ksplice.com/api/"
API_VERSION     = 1
API_CONFIG_FILE = "/etc/uptrack-api.conf"

class Client(object):
    """ A client for the Ksplice Uptrack API."""
    def __init__(self, user=None, api_key=None, api_url=None):
        """
        Accepts a username and API key for authentication.

        If the username or api key are not provided, they will be
        retrieved from the environment variables
        `UPTRACK_API_USERNAME` and `UPTRACK_API_KEY`.

        Finally, if a username and API key still cannot be found, the
        file `/etc/uptrack-api.conf` will be loaded. This file should
        be of the form:

        [uptrack]
        username = joeuser
        api_key  = 3af3c2c1ec407feb0fdc9fc1d8c4460c

        Your username and API key can be retrieved from
        <https://uptrack.ksplice.com/settings>.
        """
        if user is None:
            user = os.environ.get('UPTRACK_API_USERNAME')
        if api_key is None:
            api_key = os.environ.get('UPTRACK_API_KEY')
        if api_url is None:
            api_url = os.environ.get('UPTRACK_API_URL', None)

        config = ConfigParser.SafeConfigParser()
        try:
            config.read([API_CONFIG_FILE])
            if user is None and config.has_option('uptrack', 'username'):
                user = config.get('uptrack', 'username')
            if api_key is None and config.has_option('uptrack', 'api_key'):
                api_key = config.get('uptrack', 'api_key')
            if api_url is None and config.has_option('uptrack', 'api_url'):
                api_url = config.get('uptrack', 'api_url')
        except ConfigParser.Error, e:
            raise UsageError(e.message)

        if user is None or api_key is None:
            raise UsageError("You must provide an API username and key.")

        if api_url is None:
            api_url = API_URL

        self.user = user
        self.api_key = api_key

        parsed = urlparse.urlsplit(api_url)
        self.use_ssl = (parsed[0] == 'https')
        netloc = parsed[1]
        if ":" in netloc:
            (self.host, port) = netloc.split(':')
            self.port = int(port)
        else:
            self.host = netloc
            if self.use_ssl:
                self.port = 443
            else:
                self.port = 80

        self.api_path = parsed[2]

        self.http_connect()

    def http_connect(self):
        """Open an HTTP connection to the API host.

        This method is internal to the Client implementation.
        """
        if self.use_ssl:
            cls = httplib.HTTPSConnection
        else:
            cls = httplib.HTTPConnection
        self.connection = cls(self.host, self.port)

    def request(self, method, path, data=None):
        """Make an HTTP request to the Uptrack API.

        `method` - The HTTP method to use
        `path`   - The path on the server to request
        `data`   - Data to include (JSON-encoded) as the request body

        On a successful request (HTTP code 2xx), returns the
        JSON-decoded body of the server's response, or raises an
        UptrackError on error.
        """
        full_path = posixpath.join(self.api_path,
                                   str(API_VERSION),
                                   path)

        if data is not None:
            body = cjson.encode(data)
        else:
            body = ''

        headers = {
            'X-Uptrack-User': self.user,
            'X-Uptrack-Key' : self.api_key,
            'Content-Type'  : 'application/json'
            }

        def do_request():
            self.connection.request(method, full_path, body, headers)

        try:
            do_request()
        except httplib.ImproperConnectionState:
            self.http_connect()
            do_request()

        response = self.connection.getresponse()
        if response.status < 200 or response.status > 299:
            error = response.read()
            if response.status == 403:
                raise AuthenticationError(403, error)
            else:
                raise UptrackError(response.status, error)

        body = response.read()
        if len(body):
            return cjson.decode(body)
        return None

    def list_machines(self):
        """ List all machines known by the Uptrack service.

        Returns a list of dictionaries of the form:

         {
            hostname: uptrack.ksplice.com,
            ip: 184.73.248.238,
            last_seen: '2010-04-26T18:03:43Z',
            uuid: e82ba0ae-ad0a-4b92-a776-62b502bfd29d
            status: uptodate,
            active: true
         }

        `status` will be one of `unsupported`, `outofdate`, and `uptodate`.
        """
        return self.request('GET', 'machines')

    def authorize_machine(self, uuid):
        """ Authorize a specific UUID to access the Uptrack service.

        If you enable the "manual authorization" setting in the
        Uptrack service, the service will accept requests using your
        access key only from machines whose UUID you have specifically
        authorized.  This API call authorizes a machine for the
        Uptrack service.

        You can enable the "manual authorization" setting at
        https://uptrack.ksplice.com/settings/.

        The Uptrack UUID of a machine is stored at the path
        /var/lib/uptrack/uuid.
        """
        self.request('POST',
                     posixpath.join('machine', uuid, 'authorize'),
                     {'authorized': True})

    def deauthorize_machine(self, uuid):
        """ Revoke a machine's access to the Uptrack service.

        See `authorize_machine` for more information.
        """
        self.request('POST',
                     posixpath.join('machine', uuid, 'authorize'),
                     {'authorized': False})
